/*----- PROTECTED REGION ID(PduMapper.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        PduMapper.cpp
//
// description : C++ source for the PduMapper class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               PduMapper are implemented in this file.
//
// project :     
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
// $Author: alessandro.olivo $
//
// $Revision: 1.4 $
// $Date: 2017-11-02 09:40:30 $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <PduMapper.h>
#include <PduMapperClass.h>
#include <algorithm>
#include <iomanip>


/*----- PROTECTED REGION END -----*/	//	PduMapper.cpp

/**
 *  PduMapper class description:
 *    
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name                     |  Method name
//================================================================
//  State                            |  dev_state
//  Status                           |  dev_status
//  GetPduListFromNamePattern        |  get_pdu_list_from_name_pattern
//  GetPduListFromOutletPattern      |  get_pdu_list_from_outlet_pattern
//  GetPduInfoFromName               |  get_pdu_info_from_name
//  GetOutletsListFromOutletPattern  |  get_outlets_list_from_outlet_pattern
//================================================================

//================================================================
//  Attributes managed is:
//================================================================
//================================================================

namespace PduMapper_ns
{
/*----- PROTECTED REGION ID(PduMapper::namespace_starting) ENABLED START -----*/

//	static initializations

/*----- PROTECTED REGION END -----*/	//	PduMapper::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : PduMapper::PduMapper()
 *	Description : Constructors for a Tango device
 *                implementing the classPduMapper
 */
//--------------------------------------------------------
PduMapper::PduMapper(Tango::DeviceClass *cl, string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(PduMapper::constructor_1) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::constructor_1
}
//--------------------------------------------------------
PduMapper::PduMapper(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(PduMapper::constructor_2) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::constructor_2
}
//--------------------------------------------------------
PduMapper::PduMapper(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(PduMapper::constructor_3) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : PduMapper::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void PduMapper::delete_device()
{
	DEBUG_STREAM << "PduMapper::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(PduMapper::delete_device) ENABLED START -----*/
	
	//	Delete device allocated objects
	vector<int> events_id;
	Tango::DeviceProxy *pdu_proxy;
	Tango::DevFailed *pdu_error;
	size_t pdu_num = pdu_map->getPduNum();

	for(size_t i = 0 ; i < pdu_num; i++)
	{
		pdu_map->getAllPduEvents(events_id, i);
		pdu_proxy = pdu_map->getPduProxy(i);
		pdu_error = pdu_map->getPduErrors(i);
		size_t events_num = events_id.size();
		for(size_t j = 0; j < events_num; j ++)
		{			
			pdu_proxy->unsubscribe_event(events_id[j]);
			DEBUG_STREAM << "Unsubscribe event: " << events_id[j] << " from " << pdu_proxy->dev_name() << endl;
		}
		DEBUG_STREAM << "Delete proxy end error: " << pdu_proxy->dev_name() << endl;
		delete pdu_proxy;
		if (pdu_error)
			delete pdu_error;
	}
	
	if (dbTangoProxy)
		delete dbTangoProxy;
	
	if (attrChngCllBck)
		delete attrChngCllBck;
	
	if (pdu_map)
		PduMap::resetInstance();
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::delete_device
}

//--------------------------------------------------------
/**
 *	Method      : PduMapper::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void PduMapper::init_device()
{
	DEBUG_STREAM << "PduMapper::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(PduMapper::init_device_before) ENABLED START -----*/
	
	//	Initialization before get_device_property() call
	
	dbTangoProxy = new Tango::DeviceProxy("sys/database/2");
	attrChngCllBck = new AttributeChangeCallBack();
	pdu_map = PduMap::getInstance();
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::init_device_before
	

	//	Get the device properties from database
	get_device_property();
	
	//	No longer if mandatory property not set. 
	if (mandatoryNotDefined)
		return;

	/*----- PROTECTED REGION ID(PduMapper::init_device) ENABLED START -----*/
	
	//	Initialize device
	vector<string> pdu_list;
	unmapped_pdu_ls = "";
	set_state(Tango::UNKNOWN);
	
	try
	{
		createPduList(pdu_list);
	}
	catch(Tango::DevFailed &e)
	{
			stringstream descr;
			descr << "Device server can't create the pdu map."
				  << "Error occurred. " << endl
				  << "Reason: " << e.errors[0].reason << endl
				  << "Description: " << e.errors[0].desc << endl
				  << "Origin: " << e.errors[0].origin;

			set_state(Tango::FAULT);
			set_status(descr.str());
			return;
	}
	
	tangodb_pdu_count = pdu_list.size();
	for(unsigned int i = 0; i < tangodb_pdu_count; i++)
	{
		try
		{
			addPduData(pdu_list[i]);
		}
		catch(Tango::DevFailed &e)
		{
			unmapped_pdu_ls += "- " + pdu_list[i] + "\n";
			ERROR_STREAM << "PduMapper::init_device() create device "
					<< device_name << endl << "Reason: "
					<< e.errors[0].reason << endl
					<< "Description: " << e.errors[0].desc << endl
					<< "Origin: " << e.errors[0].origin;
		}
		catch(MapException &e)
		{
			ERROR_STREAM << "PduMapper::init_device() create device "
					<< device_name << " - " << e.what();
			unmapped_pdu_ls += "- " + pdu_list[i] + "\n";
		}
		
		
	}
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::init_device
}

//--------------------------------------------------------
/**
 *	Method      : PduMapper::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void PduMapper::get_device_property()
{
	/*----- PROTECTED REGION ID(PduMapper::get_device_property_before) ENABLED START -----*/
	
	//	Initialize property data members
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::get_device_property_before

	mandatoryNotDefined = false;

	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("pduListPattern"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on PduMapperClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		PduMapperClass	*ds_class =
			(static_cast<PduMapperClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize pduListPattern from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  pduListPattern;
		else {
			//	Try to initialize pduListPattern from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  pduListPattern;
		}
		//	And try to extract pduListPattern value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  pduListPattern;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

	}

	/*----- PROTECTED REGION ID(PduMapper::get_device_property_after) ENABLED START -----*/
	
	//	Check device property data members init
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::get_device_property_after
}
//--------------------------------------------------------
/**
 *	Method      : PduMapper::check_mandatory_property()
 *	Description : For mandatory properties check if defined in database.
 */
//--------------------------------------------------------
void PduMapper::check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop)
{
	//	Check if all properties are empty
	if (class_prop.is_empty() && dev_prop.is_empty())
	{
		TangoSys_OMemStream	tms;
		tms << endl <<"Property \'" << dev_prop.name;
		if (Tango::Util::instance()->_UseDb==true)
			tms << "\' is mandatory but not defined in database";
		else
			tms << "\' is mandatory but cannot be defined without database";
		string	status(get_status());
		status += tms.str();
		set_status(status);
		mandatoryNotDefined = true;
		/*----- PROTECTED REGION ID(PduMapper::check_mandatory_property) ENABLED START -----*/
		cerr << tms.str() << " for " << device_name << endl;
		
		/*----- PROTECTED REGION END -----*/	//	PduMapper::check_mandatory_property
	}
}


//--------------------------------------------------------
/**
 *	Method      : PduMapper::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void PduMapper::always_executed_hook()
{
	DEBUG_STREAM << "PduMapper::always_executed_hook()  " << device_name << endl;
	if (mandatoryNotDefined)
	{
		string	status(get_status());
		Tango::Except::throw_exception(
					(const char *)"PROPERTY_NOT_SET",
					status.c_str(),
					(const char *)"PduMapper::always_executed_hook()");
	}
	/*----- PROTECTED REGION ID(PduMapper::always_executed_hook) ENABLED START -----*/
	
	//	code always executed before all requests
	
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : PduMapper::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void PduMapper::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "PduMapper::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(PduMapper::read_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::read_attr_hardware
}


//--------------------------------------------------------
/**
 *	Method      : PduMapper::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void PduMapper::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(PduMapper::add_dynamic_attributes) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic attributes if any
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command State related method
 *	Description: This command gets the device state (stored in its device_state data member) and returns it to the caller.
 *
 *	@returns Device state
 */
//--------------------------------------------------------
Tango::DevState PduMapper::dev_state()
{
	DEBUG_STREAM << "PduMapper::State()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(PduMapper::dev_state) ENABLED START -----*/
	
	//	Add your own code
	Tango::DevState	argout = get_state();	
	
	if (argout != Tango::FAULT)
	{
		pdu_map->mutex.lock();
		try
		{
			if(!pdu_map->inError() and unmapped_pdu_ls == "")
				argout = Tango::ON;
			else
				argout = Tango::UNKNOWN;
			pdu_map->mutex.unlock();
		}
		catch(MapException &e)
		{
			pdu_map->mutex.unlock();
			ERROR_STREAM << "PduMapper::Status() - "
					 << device_name << " - " << e.what();
			Tango::Except::throw_exception(
				e.error_code,
				e.what(),
				"PduMapper::Status()");
		}
	}

	/*----- PROTECTED REGION END -----*/	//	PduMapper::dev_state
	set_state(argout);    // Give the state to Tango.
	if (argout!=Tango::ALARM)
		DeviceImpl::dev_state();
	return get_state();  // Return it after Tango management.
}
//--------------------------------------------------------
/**
 *	Command Status related method
 *	Description: This command gets the device status (stored in its device_status data member) and returns it to the caller.
 *
 *	@returns Device status
 */
//--------------------------------------------------------
Tango::ConstDevString PduMapper::dev_status()
{
	DEBUG_STREAM << "PduMapper::Status()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(PduMapper::dev_status) ENABLED START -----*/
	
	//	Add your own code
	string pdu_in_err_ls;
	string status = get_status();
	
	if (get_state() != Tango::FAULT)
	{
		stringstream status_ss;
		size_t pdu_count;
		pdu_map->mutex.lock();
		try
		{
			pdu_map->getPduInErrorList(pdu_in_err_ls);
			pdu_count = pdu_map->getPduNum();
			pdu_map->mutex.unlock();
		}
		catch(MapException &e)
		{
			pdu_map->mutex.unlock();
			ERROR_STREAM << "PduMapper::Status() - "
					 << device_name << " - " << e.what();
			Tango::Except::throw_exception(
				e.error_code,
				e.what(),
				"PduMapper::Status()");
		}
		if(pdu_in_err_ls == "" and  unmapped_pdu_ls == "")
		{
			status_ss << pdu_count << " of " << tangodb_pdu_count 		
					  << " PDU device server mapped and responsive" << endl;
		}
		else
		{
			status_ss << "The device server doesn't know the current "
						 "and complete PDU device servers situation." 
					  << endl << pdu_count << " of " << tangodb_pdu_count 
					  << " PDU device servers mapped." << endl
					  << ((unmapped_pdu_ls != "") ?
					  "The following PDU device servers are NOT mapped: \n" 
					  + unmapped_pdu_ls : "") 
					  << ((pdu_in_err_ls != "") ?
					  "The following PDU device servers responds with errors: \n" 
					  + pdu_in_err_ls : "");
		}
		status = status_ss.str();
	}
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::dev_status
	set_status(status);               // Give the status to Tango.
	return DeviceImpl::dev_status();  // Return it.
}
//--------------------------------------------------------
/**
 *	Command GetPduListFromNamePattern related method
 *	Description: 
 *
 *	@param argin Regex pattern, that indentify part of the Tango name of a pdu (es. uh/pdu/.* )
 *	@returns List of Tango pdu names matched (could return one or zero pdu name)
 */
//--------------------------------------------------------
Tango::DevVarStringArray *PduMapper::get_pdu_list_from_name_pattern(Tango::DevString argin)
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "PduMapper::GetPduListFromNamePattern()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(PduMapper::get_pdu_list_from_name_pattern) ENABLED START -----*/
	
	//	Add your own code
	stringstream ssname;
	string name_pattern;
        
	ssname << argin << flush;
	name_pattern = ssname.str();
        
	vector<string> pdu_ls;
	argout = new Tango::DevVarStringArray();
	
	pdu_map->mutex.lock();
	try
	{
		pdu_map->getPduListFromName(pdu_ls, name_pattern);
		pdu_map->mutex.unlock();
	}
	catch(MapException &e)
	{
		pdu_map->mutex.unlock();
		ERROR_STREAM << "PduMapper::get_pdu_list_from_name_pattern() - "
					 << device_name << " - " << e.what();
		
		delete argout;
		Tango::Except::throw_exception(
		e.error_code,
		e.what(),
		"PduMapper::get_pdu_list_from_name()");
	}
	catch(regex_error &e)
	{
		pdu_map->mutex.unlock();
		ERROR_STREAM << "PduMapper::get_pdu_list_from_name_pattern() - "
					 << device_name << " - " << e.what();
		
		delete argout;
		Tango::Except::throw_exception(
		"PduMapper::get_pdu_list_from_name_pattern()",
                "Malformed regular expression",
                e.what() + string(" code: ") + to_string(e.code()));
	}
	
	size_t len = pdu_ls.size();
	argout->length(len);
	
	for(unsigned int i = 0; i < len; i++)
		(*argout)[i] = CORBA::string_dup(pdu_ls[i].c_str());
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::get_pdu_list_from_name_pattern
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetPduListFromOutletPattern related method
 *	Description: 
 *
 *	@param argin Regex pattern, that is part of an outlet name (es. .*MSCR.* )
 *	@returns Tango name list of the pdu whose outlets match the passed argument (could return one or zero pdu name)
 */
//--------------------------------------------------------
Tango::DevVarStringArray *PduMapper::get_pdu_list_from_outlet_pattern(Tango::DevString argin)
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "PduMapper::GetPduListFromOutletPattern()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(PduMapper::get_pdu_list_from_outlet_pattern) ENABLED START -----*/
	
	//	Add your own code
	stringstream sslabel;
	string label_pattern;
        
        sslabel << argin << flush;
	label_pattern = sslabel.str();
        
	vector<string> pdu_ls;
	argout = new Tango::DevVarStringArray();
	
	pdu_map->mutex.lock();
	try
	{
		pdu_map->getPduListFromLabel(pdu_ls, label_pattern);	
		pdu_map->mutex.unlock();
	}
	catch(MapException &e)
	{
		pdu_map->mutex.unlock();
		ERROR_STREAM << "PduMapper::get_pdu_list_from_outlet_pattern() - "
					 << device_name << " - " << e.what();
		
		delete argout;
		Tango::Except::throw_exception(
		e.error_code,
		e.what(),
		"PduMapper::get_pdu_list_from_outlet_pattern()");
	}
	catch(regex_error &e)
	{
		pdu_map->mutex.unlock();
		ERROR_STREAM << "PduMapper::get_pdu_list_from_outlet_pattern() - "
					 << device_name << " - " << e.what();
		
		delete argout;
		Tango::Except::throw_exception(
		"PduMapper::get_pdu_list_from_outlet_pattern()",
                "Malformed regular expression",
                e.what() + string(" code: ") + to_string(e.code()));
	}
	
	size_t len = pdu_ls.size();
	argout->length(len);
	
	for(unsigned int i = 0; i < len; i++)
		(*argout)[i] = CORBA::string_dup(pdu_ls[i].c_str());
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::get_pdu_list_from_outlet_pattern
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetPduInfoFromName related method
 *	Description: 
 *
 *	@param argin PDU Tango name : ``domain/family/member``
 *	@returns List formatted with Name and State value of the pdu like that:
 *           [``Name01``,``State01``,``Name02``,``State02``,...]
 */
//--------------------------------------------------------
Tango::DevVarStringArray *PduMapper::get_pdu_info_from_name(Tango::DevString argin)
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "PduMapper::GetPduInfoFromName()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(PduMapper::get_pdu_info_from_name) ENABLED START -----*/
	
	//	Add your own code
	stringstream ssname;
	string name;
        
	ssname << argin << flush; 
	name = ssname.str();
        
	vector<string> pdu_info;
	argout = new Tango::DevVarStringArray();
	
	pdu_map->mutex.lock();
	try
	{
		pdu_map->getPduInfoFromName(pdu_info, name);
		pdu_map->mutex.unlock();
	}
	catch(Tango::DevFailed &e)
	{
		pdu_map->mutex.unlock();
		delete argout;
		throw e;
	}
	catch(MapException &e)
	{
		pdu_map->mutex.unlock();
		ERROR_STREAM << "PduMapper::get_pdu_info_from_name() - "
					 << device_name << " - " << e.what();
		
		delete argout;
		Tango::Except::throw_exception(
		e.error_code,
		e.what(),
		"PduMapper::get_pdu_info_from_name()");
	}
	
	size_t len = pdu_info.size();
	argout->length(len);
	
	for(unsigned int i = 0; i < len; i++)
		(*argout)[i] = CORBA::string_dup(pdu_info[i].c_str());
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::get_pdu_info_from_name
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetOutletsListFromOutletPattern related method
 *	Description: 
 *
 *	@param argin Regex pattern, that is part of an outlet name (es. .*MSCR.* )
 *	@returns List of all outlets names matched
 */
//--------------------------------------------------------
Tango::DevVarStringArray *PduMapper::get_outlets_list_from_outlet_pattern(Tango::DevString argin)
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "PduMapper::GetOutletsListFromOutletPattern()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(PduMapper::get_outlets_list_from_outlet_pattern) ENABLED START -----*/
	
	//	Add your own code
	stringstream sslabel;
	string label_pattern;
        
	sslabel << argin << flush; 
	label_pattern = sslabel.str();
        
	vector<string> outlets_ls;
	argout = new Tango::DevVarStringArray();
	
	pdu_map->mutex.lock();
	try
	{
		pdu_map->getOutletsListFromLabel(outlets_ls, label_pattern);
		pdu_map->mutex.unlock();
	}
	catch(MapException &e)
	{
		pdu_map->mutex.unlock();
		ERROR_STREAM << "PduMapper::get_outlets_list_from_outlet_pattern() - "
					 << device_name << " - " << e.what();
		delete argout;
		Tango::Except::throw_exception(
		e.error_code,
		e.what(),
		"PduMapper::get_outlets_list_from_outlet_pattern()");
	}
	catch(regex_error &e)
	{
		pdu_map->mutex.unlock();
		ERROR_STREAM << "PduMapper::get_outlets_list_from_outlet_pattern() - "
					 << device_name << " - " << e.what();
		delete argout;
		Tango::Except::throw_exception(
		"PduMapper::get_outlets_list_from_outlet_pattern()",
                "Malformed regular expression",
                e.what() + string(" code: ") + to_string(e.code()));
	}
	
	size_t len = outlets_ls.size();
	argout->length(len);
	
	for(unsigned int i = 0; i < len; i++)
		(*argout)[i] = CORBA::string_dup(outlets_ls[i].c_str());
		
	/*----- PROTECTED REGION END -----*/	//	PduMapper::get_outlets_list_from_outlet_pattern
	return argout;
}
//--------------------------------------------------------
/**
 *	Method      : PduMapper::add_dynamic_commands()
 *	Description : Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void PduMapper::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(PduMapper::add_dynamic_commands) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic commands if any
	
	/*----- PROTECTED REGION END -----*/	//	PduMapper::add_dynamic_commands
}

/*----- PROTECTED REGION ID(PduMapper::namespace_ending) ENABLED START -----*/

//	Additional Methods
// //--------------------------------------------------------
// /**
//  *	Read attribute PduList related method
//  *	Description: 
//  *
//  *	Data type:	Tango::DevString
//  *	Attr type:	Spectrum max = 300
//  */
// //--------------------------------------------------------
// void PduMapper::read_PduList(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "PduMapper::read_PduList(Tango::Attribute &attr) entering... " << endl;
// 	//	Set the attribute value
// 	attr.set_value(attr_PduList_read, 300);
// 	
// }


void PduMapper::createPduList(vector<string>& pdu_ls)
{
	DEBUG_STREAM << "PduMapper::createPduList()" << device_name << endl;
	
	Tango::DeviceData ddin, ddout;
	ddin << pduListPattern;

	ddout = dbTangoProxy->command_inout("DbGetDeviceWideList", ddin);
	ddout >> pdu_ls;
}

void PduMapper::addPduData(string dev_name)
{
	DEBUG_STREAM << "PduMapper::addPduData( " << dev_name << " ) - " << device_name << endl;
	
	unsigned short outletsNum;
	Tango::DeviceProxy *pdu_proxy = NULL;
	Tango::DeviceAttribute daout;
        
	try
	{	
		pdu_proxy = new Tango::DeviceProxy(dev_name);		
		daout = pdu_proxy->read_attribute("OutletCount");
		daout >> outletsNum;
	}
	catch(...)
	{
		if(pdu_proxy)
			delete pdu_proxy;
		throw;
	}
	try
	{
		//ADD PduData to pdu map object
		pdu_map->mutex.lock();
		pdu_map->addPduData(outletsNum, dev_name, pdu_proxy);
		pdu_map->mutex.unlock();
		
		//INIT outlet info of the PduData and SUBSCRIBE change events
		int event_id;
		for (int i = 0; i < outletsNum; i++)
		{
			
			stringstream label_attr_ss, state_attr_ss;
			string label_attr, state_attr, label, state;
			
			label_attr_ss << "name" << setfill('0') << setw(2) << i + 1 << flush;
			label_attr_ss >> label_attr;
			state_attr_ss << "state" << setfill('0') << setw(2) << i + 1 << flush;
			state_attr_ss >> state_attr;
			
			event_id = pdu_proxy->subscribe_event(label_attr, Tango::CHANGE_EVENT, attrChngCllBck);
			pdu_map->mutex.lock();
			pdu_map->updateAttrEventId(dev_name, label_attr, event_id);
			pdu_map->mutex.unlock();
			
			event_id = pdu_proxy->subscribe_event(state_attr, Tango::CHANGE_EVENT, attrChngCllBck);
			pdu_map->mutex.lock();
			pdu_map->updateAttrEventId(dev_name, state_attr, event_id);
			pdu_map->mutex.unlock();
		}
	}
	catch(...)
	{
		pdu_map->mutex.unlock();
		pdu_map->mutex.lock();
		pdu_map->deletePduData(dev_name);
		pdu_map->mutex.unlock();
		throw;
	}
}

void AttributeChangeCallBack::push_event(Tango::EventData *ev)
{
	#ifndef NDEBUG
	std::cout << "AttributeChangeCallBack::push_event()" << endl;
	#endif
	PduMap *pdu_map = PduMap::getInstance();
	Tango::DevFailed *errs = NULL;
	string pdu_name, attr_name, value;
	try
	{			
		pdu_name = ev->device->dev_name();
		
		if (!ev->err)
		{			
			attr_name = ev->attr_value->get_name();
			transform(attr_name.begin(), attr_name.end(), attr_name.begin(), ::tolower);
			
			if (attr_name.substr(0, 5).compare("state") == 0)
			{
				Tango::DevState dev_state;
				(*ev->attr_value) >> dev_state;
				value = Tango::DevStateName[dev_state];
				
				pdu_map->mutex.lock();
				pdu_map->updateOutletState(pdu_name, attr_name, value);
				pdu_map->mutex.unlock();
			}
			else if (attr_name.substr(0, 4).compare("name") == 0)
			{
				(*ev->attr_value) >> value;
				pdu_map->mutex.lock();
				pdu_map->updateOutletLabel(pdu_name, attr_name, value);
				pdu_map->mutex.unlock();
			}
			#ifndef NDEBUG
			std::cout << "Value: " << (*ev->attr_value) << endl;
			#endif
		}
		else
			errs = new Tango::DevFailed(ev->errors);
		
		pdu_map->mutex.lock();
		pdu_map->updatePduError(pdu_name, errs);
		pdu_map->mutex.unlock();
	}
	catch(Tango::DevFailed &e)
	{
		pdu_map->mutex.unlock();
		if (errs)
			delete errs;
		#ifndef NDEBUG
		std::cout << "AttributeChangeCallBack::push_event(): could not extract data !" << endl;
		#endif
		Tango::Except::print_exception(*errs);
	}
	
	#ifndef NDEBUG
	pdu_map->mutex.lock();
	std::cout << endl << pdu_map->toString() << endl;
	pdu_map->mutex.unlock();
	#endif
}


/*----- PROTECTED REGION END -----*/	//	PduMapper::namespace_ending
} //	namespace
